from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.comment import Comment
from ...arctrl_core.data import Data
from ...arctrl_core.data_context import (DataContext__get_Explication, DataContext__get_Unit, DataContext__get_ObjectType, DataContext__get_GeneratedBy, DataContext__get_Description, DataContext__get_Label, DataContext, DataContext__ctor_Z780A8A2A)
from ...arctrl_core.data_file import (DataFile__get_AsString, DataFile, DataFile_fromString_Z721C83C5)
from ...arctrl_core.Helper.collections_ import (Option_fromValueWithDefault, Option_fromSeq, ResizeArray_map)
from ...arctrl_core.Helper.regex import ActivePatterns__007CRegex_007C__007C
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Table.composite_cell import CompositeCell
from ...arctrl_core.Table.composite_header import (CompositeHeader, IOType)
from ...arctrl_core.value import Value as Value_1
from ...arctrl_file_system.file_system import FileSystem
from ...arctrl_file_system.file_system_tree import FileSystemTree
from ...arctrl_file_system.path import combine
from ...arctrl_rocrate.ldcontext import LDContext
from ...arctrl_rocrate.ldobject import (LDNode, LDRef, LDGraph)
from ...arctrl_rocrate.LDTypes.comment import LDComment
from ...arctrl_rocrate.LDTypes.dataset import LDDataset
from ...arctrl_rocrate.LDTypes.defined_term import LDDefinedTerm
from ...arctrl_rocrate.LDTypes.file import LDFile
from ...arctrl_rocrate.LDTypes.property_value import LDPropertyValue
from ...arctrl_rocrate.LDTypes.sample import LDSample
from ...fable_library.array_ import map as map_1
from ...fable_library.option import (map, default_arg, bind)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.reg_exp import (get_item, groups)
from ...fable_library.string_ import (to_fail, printf)
from ...fable_library.types import (Array, to_string)

def _expr3874() -> TypeInfo:
    return class_type("ARCtrl.Conversion.BaseTypes", None, BaseTypes)


class BaseTypes:
    ...

BaseTypes_reflection = _expr3874

def BaseTypes_composeComment_Z13201A7E(comment: Comment) -> LDNode:
    name: str
    match_value: str | None = comment.Name
    if match_value is None:
        raise Exception("Comment must have a name")

    else: 
        name = match_value

    return LDComment.create(name, None, comment.Value)


def BaseTypes_decomposeComment_Z2F770004(comment: LDNode, context: LDContext | None=None) -> Comment:
    return Comment(LDComment.get_name_as_string(comment, context), LDComment.try_get_text_as_string(comment, context))


def BaseTypes_ontologyTermFromNameAndID_40457300(name: str | None=None, id: str | None=None) -> OntologyAnnotation:
    if id is None:
        return OntologyAnnotation.create(name)

    else: 
        t: str = id
        return OntologyAnnotation.from_term_annotation(t, name)



def BaseTypes_tryOntologyTermFromNameAndID_40457300(name: str | None=None, id: str | None=None) -> OntologyAnnotation | None:
    if (id is None) if (name is None) else False:
        return None

    else: 
        return BaseTypes_ontologyTermFromNameAndID_40457300(name, id)



def BaseTypes_composeDefinedTerm_ZDED3A0F(term: OntologyAnnotation) -> LDNode:
    tan: str | None = Option_fromValueWithDefault("", term.TermAccessionAndOntobeeUrlIfShort)
    return LDDefinedTerm.create(term.NameText, None, tan)


def BaseTypes_decomposeDefinedTerm_Z2F770004(term: LDNode, context: LDContext | None=None) -> OntologyAnnotation:
    return BaseTypes_ontologyTermFromNameAndID_40457300(LDDefinedTerm.get_name_as_string(term, context), LDDefinedTerm.try_get_term_code_as_string(term, context))


def BaseTypes_composePropertyValueFromOA_ZDED3A0F(term: OntologyAnnotation) -> LDNode:
    tan: str | None = Option_fromValueWithDefault("", term.TermAccessionAndOntobeeUrlIfShort)
    return LDPropertyValue.create(term.NameText, None, None, tan)


def BaseTypes_decomposePropertyValueToOA_Z2F770004(term: LDNode, context: LDContext | None=None) -> OntologyAnnotation:
    return BaseTypes_ontologyTermFromNameAndID_40457300(LDPropertyValue.get_name_as_string(term, context), LDPropertyValue.try_get_property_idas_string(term, context))


def BaseTypes_valuesOfCell_Z436420FE(value: CompositeCell) -> tuple[str | None, str | None, str | None, str | None]:
    if value.tag == 0:
        if value.fields[0].is_empty():
            return (None, None, None, None)

        elif value.fields[0].TANInfo is not None:
            return (value.fields[0].Name, value.fields[0].TermAccessionAndOntobeeUrlIfShort, None, None)

        else: 
            return (value.fields[0].Name, None, None, None)


    elif value.tag == 2:
        pattern_input: tuple[str | None, str | None] = ((None, None)) if value.fields[1].is_empty() else ((value.fields[1].Name, Option_fromValueWithDefault("", value.fields[1].TermAccessionAndOntobeeUrlIfShort)))
        return (Option_fromValueWithDefault("", value.fields[0]), None, pattern_input[0], pattern_input[1])

    elif value.tag == 3:
        raise Exception("Data cell should not be parsed to isa value")

    elif value.fields[0] == "":
        return (None, None, None, None)

    else: 
        return (value.fields[0], None, None, None)



def BaseTypes_termOfHeader_6CAF647B(header: CompositeHeader) -> tuple[str, str | None]:
    (pattern_matching_result, oa) = (None, None)
    if header.tag == 0:
        pattern_matching_result = 0
        oa = header.fields[0]

    elif header.tag == 3:
        pattern_matching_result = 0
        oa = header.fields[0]

    elif header.tag == 2:
        pattern_matching_result = 0
        oa = header.fields[0]

    elif header.tag == 1:
        pattern_matching_result = 0
        oa = header.fields[0]

    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return (oa.NameText, oa.TermAccessionAndOntobeeUrlIfShort if (oa.TANInfo is not None) else None)

    elif pattern_matching_result == 1:
        return to_fail(printf("header %O should not be parsed to isa value"))(header)



def BaseTypes_composeComponent(header: CompositeHeader, value: CompositeCell) -> LDNode:
    pattern_input: tuple[str | None, str | None, str | None, str | None] = BaseTypes_valuesOfCell_Z436420FE(value)
    pattern_input_1: tuple[str, str | None] = BaseTypes_termOfHeader_6CAF647B(header)
    return LDPropertyValue.create_component(pattern_input_1[0], pattern_input[0], None, pattern_input_1[1], pattern_input[3], pattern_input[2], pattern_input[1])


def BaseTypes_composeParameterValue(header: CompositeHeader, value: CompositeCell) -> LDNode:
    pattern_input: tuple[str | None, str | None, str | None, str | None] = BaseTypes_valuesOfCell_Z436420FE(value)
    pattern_input_1: tuple[str, str | None] = BaseTypes_termOfHeader_6CAF647B(header)
    return LDPropertyValue.create_parameter_value(pattern_input_1[0], pattern_input[0], None, pattern_input_1[1], pattern_input[3], pattern_input[2], pattern_input[1])


def BaseTypes_composeFactorValue(header: CompositeHeader, value: CompositeCell) -> LDNode:
    pattern_input: tuple[str | None, str | None, str | None, str | None] = BaseTypes_valuesOfCell_Z436420FE(value)
    pattern_input_1: tuple[str, str | None] = BaseTypes_termOfHeader_6CAF647B(header)
    return LDPropertyValue.create_factor_value(pattern_input_1[0], pattern_input[0], None, pattern_input_1[1], pattern_input[3], pattern_input[2], pattern_input[1])


def BaseTypes_composeCharacteristicValue(header: CompositeHeader, value: CompositeCell) -> LDNode:
    pattern_input: tuple[str | None, str | None, str | None, str | None] = BaseTypes_valuesOfCell_Z436420FE(value)
    pattern_input_1: tuple[str, str | None] = BaseTypes_termOfHeader_6CAF647B(header)
    return LDPropertyValue.create_characteristic_value(pattern_input_1[0], pattern_input[0], None, pattern_input_1[1], pattern_input[3], pattern_input[2], pattern_input[1])


def BaseTypes_composeFreetextMaterialName(header_ft: str, name: str) -> str:
    return ((("" + header_ft) + "=") + name) + ""


def BaseTypes_composeFile_6CE21C7D(d: Data, fs: FileSystem | None=None) -> LDNode:
    def create_file(__unit: None=None, d: Any=d, fs: Any=fs) -> LDNode:
        data_type: str | None = map(DataFile__get_AsString, d.DataType)
        return LDFile.create(d.NameText, d.NameText, None, data_type, d.Format, d.SelectorFormat)

    if fs is None:
        return create_file(None)

    else: 
        fs_1: FileSystem = fs
        match_value: FileSystemTree | None = fs_1.Tree.TryGetPath(d.NameText)
        if match_value is not None:
            if match_value.tag == 1:
                fs_2: FileSystemTree = match_value
                file: LDNode = create_file(None)
                file.SchemaType = [LDFile.schema_type(), LDDataset.schema_type()]
                def mapping_1(fp: str, d: Any=d, fs: Any=fs) -> LDNode:
                    full_path: str = combine(d.NameText, fp)
                    return LDFile.create(full_path, full_path)

                sub_files: Array[LDNode] = list(map_1(mapping_1, fs_2.ToFilePaths(True), None))
                LDDataset.set_has_parts(file, sub_files)
                return file

            else: 
                return create_file(None)


        else: 
            return create_file(None)




def BaseTypes_decomposeFile_Z2F770004(f: LDNode, context: LDContext | None=None) -> Data:
    def mapping(dt: str, f: Any=f, context: Any=context) -> DataFile:
        return DataFile_fromString_Z721C83C5(dt)

    data_type: DataFile | None = map(mapping, LDFile.try_get_disambiguating_description_as_string(f, context))
    format: str | None = LDFile.try_get_encoding_format_as_string(f, context)
    selector_format: str | None = LDFile.try_get_usage_info_as_string(f, context)
    return Data(None, LDFile.get_name_as_string(f, context), data_type, format, selector_format)


def BaseTypes_composeFragmentDescriptor_Z4C0BEF62(dc: DataContext) -> LDNode:
    if dc.Name is None:
        raise Exception("RO-Crate parsing of DataContext failed: Cannot create a fragment descriptor without a name.")

    id: str = LDPropertyValue.gen_id_fragment_descriptor(dc.NameText)
    def mapping(e: OntologyAnnotation, dc: Any=dc) -> tuple[str | None, str | None]:
        return (e.Name, Option_fromValueWithDefault("", e.TermAccessionAndOntobeeUrlIfShort))

    pattern_input: tuple[str | None, str | None] = default_arg(map(mapping, DataContext__get_Explication(dc)), (None, None))
    def mapping_1(u: OntologyAnnotation, dc: Any=dc) -> tuple[str | None, str | None]:
        return (u.Name, Option_fromValueWithDefault("", u.TermAccessionAndOntobeeUrlIfShort))

    pattern_input_1: tuple[str | None, str | None] = default_arg(map(mapping_1, DataContext__get_Unit(dc)), (None, None))
    def f(c: Comment, dc: Any=dc) -> str:
        return to_string(c)

    disambiguating_descriptions: Array[str] | None = Option_fromSeq(ResizeArray_map(f, dc.Comments))
    data_fragment: LDNode = BaseTypes_composeFile_6CE21C7D(dc)
    def mapping_2(term: OntologyAnnotation, dc: Any=dc) -> LDNode:
        return BaseTypes_composeDefinedTerm_ZDED3A0F(term)

    pattern: LDNode | None = map(mapping_2, DataContext__get_ObjectType(dc))
    data_fragment.SetProperty(LDFile.about(), LDRef(id))
    data_fragment.SetOptionalProperty(LDFile.pattern(), pattern)
    return LDPropertyValue.create_fragment_descriptor(dc.NameText, pattern_input[0], None, pattern_input_1[1], pattern_input_1[0], pattern_input[1], DataContext__get_GeneratedBy(dc), DataContext__get_Description(dc), DataContext__get_Label(dc), disambiguating_descriptions, data_fragment)


def BaseTypes_decomposeFragmentDescriptor_Z6839B9E8(fd: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> DataContext:
    file: LDNode | None = LDPropertyValue.try_get_subject_of(fd, graph, context)
    name: str
    if file is None:
        raise Exception("RO-Crate parsing of DataContext failed: Cannot decompose a fragment descriptor without a name.")

    else: 
        f: LDNode = file
        name = LDFile.get_name_as_string(f, context)

    def mapping(pa: LDNode, fd: Any=fd, graph: Any=graph, context: Any=context) -> OntologyAnnotation:
        return BaseTypes_decomposeDefinedTerm_Z2F770004(pa, context)

    def binder(f_1: LDNode, fd: Any=fd, graph: Any=graph, context: Any=context) -> LDNode | None:
        return LDFile.try_get_pattern_as_defined_term(f_1, graph, context)

    object_type: OntologyAnnotation | None = map(mapping, bind(binder, file))
    def binder_1(f_2: LDNode, fd: Any=fd, graph: Any=graph, context: Any=context) -> str | None:
        return LDFile.try_get_encoding_format_as_string(f_2, context)

    format: str | None = bind(binder_1, file)
    def binder_2(f_3: LDNode, fd: Any=fd, graph: Any=graph, context: Any=context) -> str | None:
        return LDFile.try_get_usage_info_as_string(f_3, context)

    selector_format: str | None = bind(binder_2, file)
    explication: OntologyAnnotation | None = BaseTypes_tryOntologyTermFromNameAndID_40457300(LDPropertyValue.try_get_value_as_string(fd), LDPropertyValue.try_get_value_reference_as_string(fd))
    unit: OntologyAnnotation | None = BaseTypes_tryOntologyTermFromNameAndID_40457300(LDPropertyValue.try_get_unit_text_as_string(fd), LDPropertyValue.try_get_unit_code_as_string(fd))
    generated_by: str | None = LDPropertyValue.try_get_measurement_method_as_string(fd)
    description: str | None = LDPropertyValue.try_get_description_as_string(fd)
    def f_4(s: str, fd: Any=fd, graph: Any=graph, context: Any=context) -> Comment:
        return Comment.from_string(s)

    return DataContext__ctor_Z780A8A2A(None, name, None, format, selector_format, explication, unit, object_type, LDPropertyValue.try_get_alternate_name_as_string(fd), description, generated_by, ResizeArray_map(f_4, LDPropertyValue.get_disambiguating_descriptions_as_string(fd)))


def BaseTypes_composeProcessInput(header: CompositeHeader, value: CompositeCell, fs: FileSystem | None=None) -> LDNode:
    if header.tag == 11:
        if header.fields[0].tag == 1:
            return LDSample.create_sample(value.AsFreeText)

        elif header.fields[0].tag == 3:
            return LDSample.create_material(value.AsFreeText)

        elif header.fields[0].tag == 2:
            if value.tag == 1:
                ft: str = value.fields[0]
                return LDFile.create(ft, ft)

            elif value.tag == 3:
                return BaseTypes_composeFile_6CE21C7D(value.fields[0], fs)

            else: 
                return to_fail(printf("Could not parse input data %O"))(value)


        elif header.fields[0].tag == 4:
            n: LDNode = LDNode(BaseTypes_composeFreetextMaterialName(header.fields[0].fields[0], value.AsFreeText), [header.fields[0].fields[0]])
            n.SetProperty(LDSample.name(), value.AsFreeText)
            return n

        else: 
            return LDSample.create_source(value.AsFreeText)


    else: 
        return to_fail(printf("Could not parse input header %O"))(header)



def BaseTypes_composeProcessOutput(header: CompositeHeader, value: CompositeCell, fs: FileSystem | None=None) -> LDNode:
    (pattern_matching_result, ft_1) = (None, None)
    if header.tag == 12:
        if header.fields[0].tag == 1:
            pattern_matching_result = 0

        elif header.fields[0].tag == 3:
            pattern_matching_result = 1

        elif header.fields[0].tag == 2:
            pattern_matching_result = 2

        elif header.fields[0].tag == 4:
            pattern_matching_result = 3
            ft_1 = header.fields[0].fields[0]

        else: 
            pattern_matching_result = 0


    else: 
        pattern_matching_result = 4

    if pattern_matching_result == 0:
        return LDSample.create_sample(value.AsFreeText)

    elif pattern_matching_result == 1:
        return LDSample.create_material(value.AsFreeText)

    elif pattern_matching_result == 2:
        if value.tag == 1:
            ft: str = value.fields[0]
            return LDFile.create(ft, ft)

        elif value.tag == 3:
            return BaseTypes_composeFile_6CE21C7D(value.fields[0], fs)

        else: 
            return to_fail(printf("Could not parse output data %O"))(value)


    elif pattern_matching_result == 3:
        n: LDNode = LDNode(BaseTypes_composeFreetextMaterialName(ft_1, value.AsFreeText), [ft_1])
        n.SetProperty(LDSample.name(), value.AsFreeText)
        return n

    elif pattern_matching_result == 4:
        return to_fail(printf("Could not parse output header %O"))(header)



def BaseTypes_headerOntologyOfPropertyValue_Z2F770004(pv: LDNode, context: LDContext | None=None) -> OntologyAnnotation:
    n: str = LDPropertyValue.get_name_as_string(pv, context)
    match_value: str | None = LDPropertyValue.try_get_property_idas_string(pv, context)
    if match_value is None:
        return OntologyAnnotation(n)

    else: 
        n_ref: str = match_value
        return OntologyAnnotation.from_term_annotation(n_ref, n)



def BaseTypes_cellOfPropertyValue_Z2F770004(pv: LDNode, context: LDContext | None=None) -> CompositeCell:
    v: str | None = LDPropertyValue.try_get_value_as_string(pv, context)
    v_ref: str | None = LDPropertyValue.try_get_value_reference_as_string(pv, context)
    u: str | None = LDPropertyValue.try_get_unit_text_as_string(pv, context)
    u_ref: str | None = LDPropertyValue.try_get_unit_code_as_string(pv, context)
    (pattern_matching_result, vr, u_1, u_ref_1) = (None, None, None, None)
    if v_ref is None:
        if u is None:
            if u_ref is None:
                pattern_matching_result = 3

            else: 
                pattern_matching_result = 2
                u_ref_1 = u_ref


        elif u_ref is not None:
            pattern_matching_result = 2
            u_ref_1 = u_ref

        else: 
            pattern_matching_result = 1
            u_1 = u


    elif u is None:
        if u_ref is None:
            pattern_matching_result = 0
            vr = v_ref

        else: 
            pattern_matching_result = 4


    else: 
        pattern_matching_result = 4

    if pattern_matching_result == 0:
        return CompositeCell(0, OntologyAnnotation.from_term_annotation(vr, v))

    elif pattern_matching_result == 1:
        return CompositeCell(2, default_arg(v, ""), OntologyAnnotation(u_1))

    elif pattern_matching_result == 2:
        return CompositeCell(2, default_arg(v, ""), OntologyAnnotation.from_term_annotation(u_ref_1, u))

    elif pattern_matching_result == 3:
        return CompositeCell(0, OntologyAnnotation(v))

    elif pattern_matching_result == 4:
        arg: str = default_arg(v, "")
        return to_fail(printf("Could not parse value %s with unit %O and unit reference %O"))(arg)(u)(u_ref)



def BaseTypes_decomposePropertyValue_Z2F770004(pv: LDNode, context: LDContext | None=None) -> tuple[OntologyAnnotation, Value_1 | None, OntologyAnnotation | None]:
    def _arrow3875(__unit: None=None, pv: Any=pv, context: Any=context) -> Value_1 | None:
        v: str | None = LDPropertyValue.try_get_value_as_string(pv, context)
        v_ref: str | None = LDPropertyValue.try_get_value_reference_as_string(pv, context)
        if v_ref is None:
            return None if (v is None) else Value_1(3, v)

        else: 
            vr: str = v_ref
            return Value_1(0, OntologyAnnotation.from_term_annotation(vr, v))


    def _arrow3876(__unit: None=None, pv: Any=pv, context: Any=context) -> OntologyAnnotation | None:
        u: str | None = LDPropertyValue.try_get_unit_text_as_string(pv, context)
        u_ref: str | None = LDPropertyValue.try_get_unit_code_as_string(pv, context)
        (pattern_matching_result, u_1, u_ref_1) = (None, None, None)
        if u is None:
            if u_ref is None:
                pattern_matching_result = 2

            else: 
                pattern_matching_result = 1
                u_ref_1 = u_ref


        elif u_ref is not None:
            pattern_matching_result = 1
            u_ref_1 = u_ref

        else: 
            pattern_matching_result = 0
            u_1 = u

        if pattern_matching_result == 0:
            return OntologyAnnotation(u_1)

        elif pattern_matching_result == 1:
            return OntologyAnnotation.from_term_annotation(u_ref_1, u)

        elif pattern_matching_result == 2:
            return None


    return (BaseTypes_headerOntologyOfPropertyValue_Z2F770004(pv, context), _arrow3875(), _arrow3876())


def BaseTypes_decomposeComponent_Z2F770004(c: LDNode, context: LDContext | None=None) -> tuple[CompositeHeader, CompositeCell]:
    return (CompositeHeader(0, BaseTypes_headerOntologyOfPropertyValue_Z2F770004(c, context)), BaseTypes_cellOfPropertyValue_Z2F770004(c, context))


def BaseTypes_decomposeParameterValue_Z2F770004(c: LDNode, context: LDContext | None=None) -> tuple[CompositeHeader, CompositeCell]:
    return (CompositeHeader(3, BaseTypes_headerOntologyOfPropertyValue_Z2F770004(c, context)), BaseTypes_cellOfPropertyValue_Z2F770004(c, context))


def BaseTypes_decomposeFactorValue_Z2F770004(c: LDNode, context: LDContext | None=None) -> tuple[CompositeHeader, CompositeCell]:
    return (CompositeHeader(2, BaseTypes_headerOntologyOfPropertyValue_Z2F770004(c, context)), BaseTypes_cellOfPropertyValue_Z2F770004(c, context))


def BaseTypes_decomposeCharacteristicValue_Z2F770004(c: LDNode, context: LDContext | None=None) -> tuple[CompositeHeader, CompositeCell]:
    return (CompositeHeader(1, BaseTypes_headerOntologyOfPropertyValue_Z2F770004(c, context)), BaseTypes_cellOfPropertyValue_Z2F770004(c, context))


def BaseTypes_decomposeProcessInput_Z2F770004(pn: LDNode, context: LDContext | None=None) -> tuple[CompositeHeader, CompositeCell]:
    if LDSample.validate_source(pn, context):
        return (CompositeHeader(11, IOType(0)), CompositeCell(1, LDSample.get_name_as_string(pn, context)))

    elif LDSample.validate_material(pn, context):
        return (CompositeHeader(11, IOType(3)), CompositeCell(1, LDSample.get_name_as_string(pn, context)))

    elif LDSample.validate(pn, context):
        return (CompositeHeader(11, IOType(1)), CompositeCell(1, LDSample.get_name_as_string(pn, context)))

    elif LDFile.validate(pn, context):
        return (CompositeHeader(11, IOType(2)), CompositeCell(3, BaseTypes_decomposeFile_Z2F770004(pn, context)))

    else: 
        n: LDNode = pn
        return (CompositeHeader(11, IOType(4, n.SchemaType[0])), CompositeCell(1, LDSample.get_name_as_string(n, context)))



def BaseTypes_decomposeProcessOutput_Z2F770004(pn: LDNode, context: LDContext | None=None) -> tuple[CompositeHeader, CompositeCell]:
    if LDSample.validate_material(pn, context):
        return (CompositeHeader(12, IOType(3)), CompositeCell(1, LDSample.get_name_as_string(pn, context)))

    elif LDSample.validate(pn, context):
        return (CompositeHeader(12, IOType(1)), CompositeCell(1, LDSample.get_name_as_string(pn, context)))

    elif LDFile.validate(pn, context):
        return (CompositeHeader(12, IOType(2)), CompositeCell(3, BaseTypes_decomposeFile_Z2F770004(pn, context)))

    else: 
        n: LDNode = pn
        return (CompositeHeader(12, IOType(4, n.SchemaType[0])), CompositeCell(1, LDSample.get_name_as_string(n, context)))



def BaseTypes_composeTechnologyPlatform_ZDED3A0F(tp: OntologyAnnotation) -> str:
    match_value: dict[str, Any] | None = tp.TANInfo
    if match_value is None:
        return ("" + tp.NameText) + ""

    else: 
        return ((("" + tp.NameText) + " (") + tp.TermAccessionShort) + ")"



def BaseTypes_decomposeTechnologyPlatform_Z721C83C5(name: str) -> OntologyAnnotation:
    active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C("^(?<value>.+) \\((?<ontology>[^(]*:[^)]*)\\)$", name)
    if active_pattern_result is not None:
        r: Any = active_pattern_result
        oa: OntologyAnnotation
        tan: str = get_item(groups(r), "ontology") or ""
        oa = OntologyAnnotation.from_term_annotation(tan)
        v: str = get_item(groups(r), "value") or ""
        return OntologyAnnotation.create(v, oa.TermSourceREF, oa.TermAccessionNumber)

    else: 
        return OntologyAnnotation.create(name)



__all__ = ["BaseTypes_reflection", "BaseTypes_composeComment_Z13201A7E", "BaseTypes_decomposeComment_Z2F770004", "BaseTypes_ontologyTermFromNameAndID_40457300", "BaseTypes_tryOntologyTermFromNameAndID_40457300", "BaseTypes_composeDefinedTerm_ZDED3A0F", "BaseTypes_decomposeDefinedTerm_Z2F770004", "BaseTypes_composePropertyValueFromOA_ZDED3A0F", "BaseTypes_decomposePropertyValueToOA_Z2F770004", "BaseTypes_valuesOfCell_Z436420FE", "BaseTypes_termOfHeader_6CAF647B", "BaseTypes_composeComponent", "BaseTypes_composeParameterValue", "BaseTypes_composeFactorValue", "BaseTypes_composeCharacteristicValue", "BaseTypes_composeFreetextMaterialName", "BaseTypes_composeFile_6CE21C7D", "BaseTypes_decomposeFile_Z2F770004", "BaseTypes_composeFragmentDescriptor_Z4C0BEF62", "BaseTypes_decomposeFragmentDescriptor_Z6839B9E8", "BaseTypes_composeProcessInput", "BaseTypes_composeProcessOutput", "BaseTypes_headerOntologyOfPropertyValue_Z2F770004", "BaseTypes_cellOfPropertyValue_Z2F770004", "BaseTypes_decomposePropertyValue_Z2F770004", "BaseTypes_decomposeComponent_Z2F770004", "BaseTypes_decomposeParameterValue_Z2F770004", "BaseTypes_decomposeFactorValue_Z2F770004", "BaseTypes_decomposeCharacteristicValue_Z2F770004", "BaseTypes_decomposeProcessInput_Z2F770004", "BaseTypes_decomposeProcessOutput_Z2F770004", "BaseTypes_composeTechnologyPlatform_ZDED3A0F", "BaseTypes_decomposeTechnologyPlatform_Z721C83C5"]




from __future__ import annotations
from collections.abc import Callable
from math import pow
from typing import Any
from ..fable_library.char import (char_code_at, is_letter, is_digit)
from ..fable_library.int32 import parse
from ..fable_library.long import (op_addition, from_integer, to_int)
from ..fable_library.option import value as value_2
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.reg_exp import create
from ..fable_library.seq import iterate
from ..fable_library.string_ import (to_text, printf, to_fail)
from ..fable_library.system_text import (StringBuilder__ctor, StringBuilder__Append_244C7CD6)
from ..fable_library.types import (uint32, to_string, int64)
from ..fable_library.util import (ignore, string_hash, identity_hash)
from .hash_codes import merge_hashes

CellReference_indexRegex: Any = create("([A-Z]*)(\\d*)")

def CellReference_colAdressToIndex(column_adress: str) -> uint32:
    length: int = len(column_adress) or 0
    sum: uint32 = uint32(0)
    for i in range(0, (length - 1) + 1, 1):
        c_1: str
        c: str = column_adress[(length - 1) - i]
        c_1 = c.upper()
        factor: uint32
        value: float = pow(26.0, i)
        factor = int(value+0x100000000 if value < 0 else value)
        sum = sum + (((int(char_code_at(c_1, 0)+0x100000000 if char_code_at(c_1, 0) < 0 else char_code_at(c_1, 0))) - uint32(64)) * factor)
    return sum


def CellReference_indexToColAdress(i: uint32) -> str:
    def loop(index_mut: uint32, acc_mut: str, i: Any=i) -> str:
        while True:
            (index, acc) = (index_mut, acc_mut)
            if index == uint32(0):
                return acc

            else: 
                mod26: uint32 = (index - uint32(1)) % uint32(26)
                index_mut = (index - uint32(1)) // uint32(26)
                acc_mut = chr((int(char_code_at("A", 0)+0x100000000 if char_code_at("A", 0) < 0 else char_code_at("A", 0))) + mod26) + acc
                continue

            break

    return loop(i, "")


def CellReference_ofIndices(column: uint32, row: uint32) -> str:
    arg: str = CellReference_indexToColAdress(column)
    return to_text(printf("%s%i"))(arg)(row)


def CellReference_toIndices(reference: str) -> tuple[uint32, uint32]:
    char_part: Any = StringBuilder__ctor()
    num_part: Any = StringBuilder__ctor()
    def action(c: str, reference: Any=reference) -> None:
        if is_letter(c):
            ignore(StringBuilder__Append_244C7CD6(char_part, c))

        elif is_digit(c):
            ignore(StringBuilder__Append_244C7CD6(num_part, c))

        else: 
            to_fail(printf("Reference %s does not match Excel A1-style"))(reference)


    iterate(action, reference)
    return (CellReference_colAdressToIndex(to_string(char_part)), parse(to_string(num_part), 511, True, 32))


def CellReference_toColIndex(reference: str) -> uint32:
    return CellReference_toIndices(reference)[0]


def CellReference_toRowIndex(reference: str) -> uint32:
    return CellReference_toIndices(reference)[1]


def CellReference_setColIndex(col_i: uint32, reference: str) -> str:
    return CellReference_ofIndices(col_i, CellReference_toIndices(reference)[1])


def CellReference_setRowIndex(row_i: uint32, reference: str) -> str:
    return CellReference_ofIndices(CellReference_toIndices(reference)[0], row_i)


def CellReference_moveHorizontal(amount: int, reference: str) -> str:
    tupled_arg_1: tuple[uint32, uint32]
    tupled_arg: tuple[uint32, uint32] = CellReference_toIndices(reference)
    def _arrow102(__unit: None=None, amount: Any=amount, reference: Any=reference) -> uint32:
        value: int64 = op_addition(from_integer(tupled_arg[0], False, 6), from_integer(amount, False, 2))
        return int(to_int(value)+0x100000000 if to_int(value) < 0 else to_int(value))

    tupled_arg_1 = (_arrow102(), tupled_arg[1])
    return CellReference_ofIndices(tupled_arg_1[0], tupled_arg_1[1])


def CellReference_moveVertical(amount: int, reference: str) -> str:
    tupled_arg_1: tuple[uint32, uint32]
    tupled_arg: tuple[uint32, uint32] = CellReference_toIndices(reference)
    def _arrow103(__unit: None=None, amount: Any=amount, reference: Any=reference) -> uint32:
        value: int64 = op_addition(from_integer(tupled_arg[1], False, 6), from_integer(amount, False, 2))
        return int(to_int(value)+0x100000000 if to_int(value) < 0 else to_int(value))

    tupled_arg_1 = (tupled_arg[0], _arrow103())
    return CellReference_ofIndices(tupled_arg_1[0], tupled_arg_1[1])


def _expr106() -> TypeInfo:
    return class_type("FsSpreadsheet.FsAddress", None, FsAddress)


class FsAddress:
    def __init__(self, row_number: int, column_number: int, fixed_row: bool | None=None, fixed_column: bool | None=None) -> None:
        self._fixedRow: bool = value_2(fixed_row) if (fixed_row is not None) else False
        self._fixedColumn: bool = value_2(fixed_column) if (fixed_column is not None) else False
        self._rowNumber: int = row_number or 0
        self._columnNumber: int = column_number or 0
        self._trimmedAddress: str = ""

    @staticmethod
    def from_string(cell_address_string: str, fixed_row: bool | None=None, fixed_column: bool | None=None) -> FsAddress:
        pattern_input: tuple[uint32, uint32] = CellReference_toIndices(cell_address_string)
        return FsAddress(int(pattern_input[1]), int(pattern_input[0]), fixed_row, fixed_column)

    @property
    def ColumnNumber(self, __unit: None=None) -> int:
        self_1: FsAddress = self
        return self_1._columnNumber

    @ColumnNumber.setter
    def ColumnNumber(self, col_i: int) -> None:
        self_1: FsAddress = self
        self_1._columnNumber = col_i or 0

    @property
    def RowNumber(self, __unit: None=None) -> int:
        self_1: FsAddress = self
        return self_1._rowNumber

    @RowNumber.setter
    def RowNumber(self, row_i: int) -> None:
        self_1: FsAddress = self
        self_1._rowNumber = row_i or 0

    @property
    def Address(self, __unit: None=None) -> str:
        self_1: FsAddress = self
        return CellReference_ofIndices(int(self_1._columnNumber+0x100000000 if self_1._columnNumber < 0 else self_1._columnNumber), int(self_1._rowNumber+0x100000000 if self_1._rowNumber < 0 else self_1._rowNumber))

    @Address.setter
    def Address(self, address: str) -> None:
        self_1: FsAddress = self
        pattern_input: tuple[uint32, uint32] = CellReference_toIndices(address)
        self_1._rowNumber = int(pattern_input[1]) or 0
        self_1._columnNumber = int(pattern_input[0]) or 0

    @property
    def FixedRow(self, __unit: None=None) -> bool:
        return False

    @property
    def FixedColumn(self, __unit: None=None) -> bool:
        return False

    def Copy(self, __unit: None=None) -> FsAddress:
        this: FsAddress = self
        col_no: int = this.ColumnNumber or 0
        return FsAddress(this.RowNumber, col_no, this.FixedRow, this.FixedColumn)

    @staticmethod
    def copy(address: FsAddress) -> FsAddress:
        return address.Copy()

    def UpdateIndices(self, row_index: int, col_index: int) -> None:
        self_1: FsAddress = self
        self_1._columnNumber = col_index or 0
        self_1._rowNumber = row_index or 0

    @staticmethod
    def update_indices(row_index: int, col_index: int, address: FsAddress) -> FsAddress:
        address.UpdateIndices(row_index, col_index)
        return address

    def ToIndices(self, __unit: None=None) -> tuple[int, int]:
        self_1: FsAddress = self
        return (self_1._rowNumber, self_1._columnNumber)

    @staticmethod
    def to_indices(address: FsAddress) -> tuple[int, int]:
        return address.ToIndices()

    def Compare(self, address: FsAddress) -> bool:
        self_1: FsAddress = self
        return (self_1.FixedRow == address.FixedRow) if ((self_1.FixedColumn == address.FixedColumn) if ((self_1.RowNumber == address.RowNumber) if ((self_1.ColumnNumber == address.ColumnNumber) if (self_1.Address == address.Address) else False) else False) else False) else False

    @staticmethod
    def compare(address1: FsAddress, address2: FsAddress) -> bool:
        return address1.Compare(address2)

    def __hash__(self, __unit: None=None) -> int:
        this: FsAddress = self
        hash_8: int
        hash_6: int
        hash_4: int
        hash_2: int = string_hash(this.Address) or 0
        hash_4 = merge_hashes(this.ColumnNumber, hash_2)
        hash_6 = merge_hashes(this.RowNumber, hash_4)
        def _arrow104(__unit: None=None) -> int:
            copy_of_struct: bool = this.FixedColumn
            return identity_hash(copy_of_struct)

        hash_8 = merge_hashes(_arrow104(), hash_6)
        def _arrow105(__unit: None=None) -> int:
            copy_of_struct_1: bool = this.FixedRow
            return identity_hash(copy_of_struct_1)

        return merge_hashes(_arrow105(), hash_8)

    def __eq__(self, other: Any=None) -> bool:
        this: FsAddress = self
        return this.Compare(other) if isinstance(other, FsAddress) else False


FsAddress_reflection = _expr106

def FsAddress__ctor_42410CA0(row_number: int, column_number: int, fixed_row: bool | None=None, fixed_column: bool | None=None) -> FsAddress:
    return FsAddress(row_number, column_number, fixed_row, fixed_column)


__all__ = ["CellReference_indexRegex", "CellReference_colAdressToIndex", "CellReference_indexToColAdress", "CellReference_ofIndices", "CellReference_toIndices", "CellReference_toColIndex", "CellReference_toRowIndex", "CellReference_setColIndex", "CellReference_setRowIndex", "CellReference_moveHorizontal", "CellReference_moveVertical", "FsAddress_reflection"]




from decimal import Decimal
from typing import Any

from .types import FSharpRef


def compare(x: int, y: int) -> int:
    return -1 if x < y else 1 if x > y else 0


def equals(a: int, b: int) -> bool:
    return a == b


def abs(x: int) -> int:
    return -x if x < 0 else x


def sign(x: int) -> int:
    return -1 if x < 0 else 1 if x > 0 else 0


def max(x: int, y: int) -> int:
    return x if x > y else y


def min(x: int, y: int) -> int:
    return x if x < y else y


def add(x: int, y: int) -> int:
    return x + y


def subtract(x: int, y: int) -> int:
    return x - y


def multiply(x: int, y: int) -> int:
    return x * y


def divide(x: int, y: int) -> int:
    return x // y


def remainder(x: int, y: int) -> int:
    return x % y


def negate(x: int) -> int:
    return -x


def op_unary_negation(x: int) -> int:
    return -x


def op_logical_not(x: int) -> int:
    return ~x


def op_unary_plus(x: int) -> int:
    return x


def op_addition(a: int, b: int) -> int:
    return a + b


def op_subtraction(a: int, b: int) -> int:
    return a - b


def op_multiply(a: int, b: int) -> int:
    return a * b


def op_division(a: int, b: int) -> int:
    return a // b


def op_modulus(a: int, b: int) -> int:
    return a % b


def op_right_shift(a: int, num_bits: int) -> int:
    return a >> num_bits


def op_left_shift(a: int, num_bits: int) -> int:
    return a << num_bits


def op_bitwise_and(x: int, y: int) -> int:
    return x & y


def op_bitwise_or(x: int, y: int) -> int:
    return x | y


def op_exclusive_or(x: int, y: int) -> int:
    return x ^ y


def op_less_than(x: int, y: int) -> bool:
    return x < y


def op_less_than_or_equal(x: int, y: int) -> bool:
    return x <= y


def op_greater_than(x: int, y: int) -> bool:
    return x > y


def op_greater_than_or_equal(x: int, y: int) -> bool:
    return x >= y


def op_equality(x: int, y: int) -> bool:
    return x == y


def op_inequality(x: int, y: int) -> bool:
    return x != y


def get_zero() -> int:
    return 0


def get_one() -> int:
    return 1


def get_minus_one() -> int:
    return -1


def get_is_zero(x: int) -> bool:
    return x == 0


def get_is_one(x: int) -> bool:
    return x == 1


def get_is_even(x: int) -> bool:
    return is_even_integer(x)


def get_is_power_of_two(x: int) -> bool:
    return is_pow2(x)


def get_sign(x: int) -> int:
    return sign(x)


def is_negative(x: int) -> bool:
    return x < 0


def is_positive(x: int) -> bool:
    return x > 0


def is_even_integer(x: int) -> bool:
    return x % 2 == 0


def is_odd_integer(x: int) -> bool:
    return x % 2 != 0


def is_pow2(x: int) -> bool:
    return (x & (x - 1)) == 0


def from_zero() -> int:
    return 0


def from_one() -> int:
    return 1


def from_int32(x: int) -> int:
    return int(x)


def from_int64(x: int) -> int:
    return int(x)


def from_string(x: str) -> int:
    return int(x)


def parse(value: Any) -> int:
    return int(value)


def to_string(value: int) -> str:
    return str(value)


def to_byte(value: int) -> int:
    return value


def to_sbyte(value: int) -> int:
    return value


def to_int16(value: int) -> int:
    return value


def to_uint16(value: int) -> int:
    return value


def to_int32(value: int) -> int:
    return value


def to_uint32(value: int) -> int:
    return value


def to_int64(value: int) -> int:
    return value


def to_uint64(value: int) -> int:
    return value


def to_single(value: int) -> float:
    return float(value)


def to_double(value: int) -> float:
    return float(value)


def to_decimal(value: int) -> Decimal:
    return Decimal(value)


def try_parse(string: str, def_value: FSharpRef[int]) -> bool:
    try:
        def_value.contents = parse(string)
        return True
    except Exception:
        return False


BigInteger = int

__all__ = [
    "BigInteger",
    "compare",
    "equals",
    "abs",
    "sign",
    "max",
    "min",
    "add",
    "subtract",
    "multiply",
    "divide",
    "remainder",
    "negate",
    "op_unary_negation",
    "op_logical_not",
    "op_unary_plus",
    "op_addition",
    "op_subtraction",
    "op_multiply",
    "op_division",
    "op_modulus",
    "op_right_shift",
    "op_left_shift",
    "op_bitwise_and",
    "op_bitwise_or",
    "op_exclusive_or",
    "op_less_than",
    "op_less_than_or_equal",
    "op_greater_than",
    "op_greater_than_or_equal",
    "op_equality",
    "op_inequality",
    "get_zero",
    "get_one",
    "get_minus_one",
    "get_is_zero",
    "get_is_one",
    "get_is_even",
    "get_is_power_of_two",
    "get_sign",
    "is_negative",
    "is_positive",
    "is_even_integer",
    "is_odd_integer",
    "is_pow2",
    "from_zero",
    "from_one",
    "from_int32",
    "from_int64",
    "from_string",
    "to_decimal",
    "to_double",
    "to_single",
    "to_uint64",
    "to_int64",
    "to_uint32",
    "to_int32",
    "to_uint16",
    "to_int16",
    "to_byte",
    "to_sbyte",
    "to_string",
    "parse",
    "try_parse",
]



from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..arctrl_core.data_map import DataMap
from ..arctrl_core.Helper.identifier import (Assay_datamapFileNameFromIdentifier, Study_datamapFileNameFromIdentifier, Workflow_datamapFileNameFromIdentifier, Run_datamapFileNameFromIdentifier)
from ..arctrl_file_system.path import combine_many
from ..arctrl_spreadsheet.data_map import (to_fs_workbook, from_fs_workbook)
from ..fable_library.array_ import equals_with
from ..fable_library.types import Array
from ..fable_library.util import safe_hash
from .contract import (Contract, DTOType, DTO)

def _007CDatamapPath_007C__007C(input: Array[str]) -> str | None:
    (pattern_matching_result,) = (None,)
    def _arrow3707(x: str, y: str, input: Any=input) -> bool:
        return x == y

    if (len(input) == 3) if (not equals_with(_arrow3707, input, None)) else False:
        if input[0] == "assays":
            if input[2] == "isa.datamap.xlsx":
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 4


        elif input[0] == "studies":
            if input[2] == "isa.datamap.xlsx":
                pattern_matching_result = 1

            else: 
                pattern_matching_result = 4


        elif input[0] == "workflows":
            if input[2] == "isa.datamap.xlsx":
                pattern_matching_result = 2

            else: 
                pattern_matching_result = 4


        elif input[0] == "runs":
            if input[2] == "isa.datamap.xlsx":
                pattern_matching_result = 3

            else: 
                pattern_matching_result = 4


        else: 
            pattern_matching_result = 4


    else: 
        pattern_matching_result = 4

    if pattern_matching_result == 0:
        any_assay_name: str = input[1]
        return combine_many(input)

    elif pattern_matching_result == 1:
        any_study_name: str = input[1]
        return combine_many(input)

    elif pattern_matching_result == 2:
        any_workflow_name: str = input[1]
        return combine_many(input)

    elif pattern_matching_result == 3:
        any_run_name: str = input[1]
        return combine_many(input)

    elif pattern_matching_result == 4:
        return None



def ARCtrl_DataMap__DataMap_ToCreateContractForAssay_Z721C83C5(this: DataMap, assay_identifier: str) -> Contract:
    path: str = Assay_datamapFileNameFromIdentifier(assay_identifier)
    return Contract.create_create(path, DTOType(5), DTO(0, to_fs_workbook(this)))


def ARCtrl_DataMap__DataMap_ToUpdateContractForAssay_Z721C83C5(this: DataMap, assay_identifier: str) -> Contract:
    path: str = Assay_datamapFileNameFromIdentifier(assay_identifier)
    return Contract.create_update(path, DTOType(5), DTO(0, to_fs_workbook(this)))


def ARCtrl_DataMap__DataMap_ToDeleteContractForAssay_Z721C83C5(this: DataMap, assay_identifier: str) -> Contract:
    path: str = Assay_datamapFileNameFromIdentifier(assay_identifier)
    return Contract.create_delete(path)


def ARCtrl_DataMap__DataMap_toDeleteContractForAssay_Static_Z721C83C5(assay_identifier: str) -> Callable[[DataMap], Contract]:
    def _arrow3708(data_map: DataMap, assay_identifier: Any=assay_identifier) -> Contract:
        return ARCtrl_DataMap__DataMap_ToDeleteContractForAssay_Z721C83C5(data_map, assay_identifier)

    return _arrow3708


def ARCtrl_DataMap__DataMap_toUpdateContractForAssay_Static_Z721C83C5(assay_identifier: str) -> Callable[[DataMap], Contract]:
    def _arrow3709(data_map: DataMap, assay_identifier: Any=assay_identifier) -> Contract:
        return ARCtrl_DataMap__DataMap_ToUpdateContractForAssay_Z721C83C5(data_map, assay_identifier)

    return _arrow3709


def ARCtrl_DataMap__DataMap_tryFromReadContractForAssay_Static(assay_identifier: str, c: Contract) -> DataMap | None:
    path: str = Assay_datamapFileNameFromIdentifier(assay_identifier)
    (pattern_matching_result, fsworkbook_1, p_1) = (None, None, None)
    if c.Operation == "READ":
        if c.DTOType is not None:
            if c.DTOType.tag == 5:
                if c.DTO is not None:
                    if c.DTO.tag == 0:
                        def _arrow3710(__unit: None=None, assay_identifier: Any=assay_identifier, c: Any=c) -> bool:
                            fsworkbook: Any = c.DTO.fields[0]
                            return c.Path == path

                        if _arrow3710():
                            pattern_matching_result = 0
                            fsworkbook_1 = c.DTO.fields[0]
                            p_1 = c.Path

                        else: 
                            pattern_matching_result = 1


                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        dm: DataMap = from_fs_workbook(fsworkbook_1)
        dm.StaticHash = safe_hash(dm) or 0
        return dm

    elif pattern_matching_result == 1:
        return None



def ARCtrl_DataMap__DataMap_ToCreateContractForStudy_Z721C83C5(this: DataMap, study_identifier: str) -> Contract:
    path: str = Study_datamapFileNameFromIdentifier(study_identifier)
    return Contract.create_create(path, DTOType(5), DTO(0, to_fs_workbook(this)))


def ARCtrl_DataMap__DataMap_ToUpdateContractForStudy_Z721C83C5(this: DataMap, study_identifier: str) -> Contract:
    path: str = Study_datamapFileNameFromIdentifier(study_identifier)
    return Contract.create_update(path, DTOType(5), DTO(0, to_fs_workbook(this)))


def ARCtrl_DataMap__DataMap_ToDeleteContractForStudy_Z721C83C5(this: DataMap, study_identifier: str) -> Contract:
    path: str = Study_datamapFileNameFromIdentifier(study_identifier)
    return Contract.create_delete(path)


def ARCtrl_DataMap__DataMap_toDeleteContractForStudy_Static_Z721C83C5(study_identifier: str) -> Callable[[DataMap], Contract]:
    def _arrow3711(data_map: DataMap, study_identifier: Any=study_identifier) -> Contract:
        return ARCtrl_DataMap__DataMap_ToDeleteContractForStudy_Z721C83C5(data_map, study_identifier)

    return _arrow3711


def ARCtrl_DataMap__DataMap_toUpdateContractForStudy_Static_Z721C83C5(study_identifier: str) -> Callable[[DataMap], Contract]:
    def _arrow3712(data_map: DataMap, study_identifier: Any=study_identifier) -> Contract:
        return ARCtrl_DataMap__DataMap_ToUpdateContractForStudy_Z721C83C5(data_map, study_identifier)

    return _arrow3712


def ARCtrl_DataMap__DataMap_tryFromReadContractForStudy_Static(study_identifier: str, c: Contract) -> DataMap | None:
    path: str = Study_datamapFileNameFromIdentifier(study_identifier)
    (pattern_matching_result, fsworkbook_1, p_1) = (None, None, None)
    if c.Operation == "READ":
        if c.DTOType is not None:
            if c.DTOType.tag == 5:
                if c.DTO is not None:
                    if c.DTO.tag == 0:
                        def _arrow3713(__unit: None=None, study_identifier: Any=study_identifier, c: Any=c) -> bool:
                            fsworkbook: Any = c.DTO.fields[0]
                            return c.Path == path

                        if _arrow3713():
                            pattern_matching_result = 0
                            fsworkbook_1 = c.DTO.fields[0]
                            p_1 = c.Path

                        else: 
                            pattern_matching_result = 1


                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        dm: DataMap = from_fs_workbook(fsworkbook_1)
        dm.StaticHash = safe_hash(dm) or 0
        return dm

    elif pattern_matching_result == 1:
        return None



def ARCtrl_DataMap__DataMap_ToCreateContractForWorkflow_Z721C83C5(this: DataMap, workflow_identifier: str) -> Contract:
    path: str = Workflow_datamapFileNameFromIdentifier(workflow_identifier)
    return Contract.create_create(path, DTOType(5), DTO(0, to_fs_workbook(this)))


def ARCtrl_DataMap__DataMap_ToUpdateContractForWorkflow_Z721C83C5(this: DataMap, workflow_identifier: str) -> Contract:
    path: str = Workflow_datamapFileNameFromIdentifier(workflow_identifier)
    return Contract.create_update(path, DTOType(5), DTO(0, to_fs_workbook(this)))


def ARCtrl_DataMap__DataMap_ToDeleteContractForWorkflow_Z721C83C5(this: DataMap, workflow_identifier: str) -> Contract:
    path: str = Workflow_datamapFileNameFromIdentifier(workflow_identifier)
    return Contract.create_delete(path)


def ARCtrl_DataMap__DataMap_toDeleteContractForWorkflow_Static_Z721C83C5(workflow_identifier: str) -> Callable[[DataMap], Contract]:
    def _arrow3717(data_map: DataMap, workflow_identifier: Any=workflow_identifier) -> Contract:
        return ARCtrl_DataMap__DataMap_ToDeleteContractForWorkflow_Z721C83C5(data_map, workflow_identifier)

    return _arrow3717


def ARCtrl_DataMap__DataMap_toUpdateContractForWorkflow_Static_Z721C83C5(workflow_identifier: str) -> Callable[[DataMap], Contract]:
    def _arrow3718(data_map: DataMap, workflow_identifier: Any=workflow_identifier) -> Contract:
        return ARCtrl_DataMap__DataMap_ToUpdateContractForWorkflow_Z721C83C5(data_map, workflow_identifier)

    return _arrow3718


def ARCtrl_DataMap__DataMap_tryFromReadContractForWorkflow_Static(workflow_identifier: str, c: Contract) -> DataMap | None:
    path: str = Workflow_datamapFileNameFromIdentifier(workflow_identifier)
    (pattern_matching_result, fsworkbook_1, p_1) = (None, None, None)
    if c.Operation == "READ":
        if c.DTOType is not None:
            if c.DTOType.tag == 5:
                if c.DTO is not None:
                    if c.DTO.tag == 0:
                        def _arrow3719(__unit: None=None, workflow_identifier: Any=workflow_identifier, c: Any=c) -> bool:
                            fsworkbook: Any = c.DTO.fields[0]
                            return c.Path == path

                        if _arrow3719():
                            pattern_matching_result = 0
                            fsworkbook_1 = c.DTO.fields[0]
                            p_1 = c.Path

                        else: 
                            pattern_matching_result = 1


                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        dm: DataMap = from_fs_workbook(fsworkbook_1)
        dm.StaticHash = safe_hash(dm) or 0
        return dm

    elif pattern_matching_result == 1:
        return None



def ARCtrl_DataMap__DataMap_ToCreateContractForRun_Z721C83C5(this: DataMap, run_identifier: str) -> Contract:
    path: str = Run_datamapFileNameFromIdentifier(run_identifier)
    return Contract.create_create(path, DTOType(5), DTO(0, to_fs_workbook(this)))


def ARCtrl_DataMap__DataMap_ToUpdateContractForRun_Z721C83C5(this: DataMap, run_identifier: str) -> Contract:
    path: str = Run_datamapFileNameFromIdentifier(run_identifier)
    return Contract.create_update(path, DTOType(5), DTO(0, to_fs_workbook(this)))


def ARCtrl_DataMap__DataMap_ToDeleteContractForRun_Z721C83C5(this: DataMap, run_identifier: str) -> Contract:
    path: str = Run_datamapFileNameFromIdentifier(run_identifier)
    return Contract.create_delete(path)


def ARCtrl_DataMap__DataMap_toDeleteContractForRun_Static_Z721C83C5(run_identifier: str) -> Callable[[DataMap], Contract]:
    def _arrow3720(data_map: DataMap, run_identifier: Any=run_identifier) -> Contract:
        return ARCtrl_DataMap__DataMap_ToDeleteContractForRun_Z721C83C5(data_map, run_identifier)

    return _arrow3720


def ARCtrl_DataMap__DataMap_toUpdateContractForRun_Static_Z721C83C5(run_identifier: str) -> Callable[[DataMap], Contract]:
    def _arrow3721(data_map: DataMap, run_identifier: Any=run_identifier) -> Contract:
        return ARCtrl_DataMap__DataMap_ToUpdateContractForRun_Z721C83C5(data_map, run_identifier)

    return _arrow3721


def ARCtrl_DataMap__DataMap_tryFromReadContractForRun_Static(run_identifier: str, c: Contract) -> DataMap | None:
    path: str = Run_datamapFileNameFromIdentifier(run_identifier)
    (pattern_matching_result, fsworkbook_1, p_1) = (None, None, None)
    if c.Operation == "READ":
        if c.DTOType is not None:
            if c.DTOType.tag == 5:
                if c.DTO is not None:
                    if c.DTO.tag == 0:
                        def _arrow3722(__unit: None=None, run_identifier: Any=run_identifier, c: Any=c) -> bool:
                            fsworkbook: Any = c.DTO.fields[0]
                            return c.Path == path

                        if _arrow3722():
                            pattern_matching_result = 0
                            fsworkbook_1 = c.DTO.fields[0]
                            p_1 = c.Path

                        else: 
                            pattern_matching_result = 1


                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        dm: DataMap = from_fs_workbook(fsworkbook_1)
        dm.StaticHash = safe_hash(dm) or 0
        return dm

    elif pattern_matching_result == 1:
        return None



__all__ = ["_007CDatamapPath_007C__007C", "ARCtrl_DataMap__DataMap_ToCreateContractForAssay_Z721C83C5", "ARCtrl_DataMap__DataMap_ToUpdateContractForAssay_Z721C83C5", "ARCtrl_DataMap__DataMap_ToDeleteContractForAssay_Z721C83C5", "ARCtrl_DataMap__DataMap_toDeleteContractForAssay_Static_Z721C83C5", "ARCtrl_DataMap__DataMap_toUpdateContractForAssay_Static_Z721C83C5", "ARCtrl_DataMap__DataMap_tryFromReadContractForAssay_Static", "ARCtrl_DataMap__DataMap_ToCreateContractForStudy_Z721C83C5", "ARCtrl_DataMap__DataMap_ToUpdateContractForStudy_Z721C83C5", "ARCtrl_DataMap__DataMap_ToDeleteContractForStudy_Z721C83C5", "ARCtrl_DataMap__DataMap_toDeleteContractForStudy_Static_Z721C83C5", "ARCtrl_DataMap__DataMap_toUpdateContractForStudy_Static_Z721C83C5", "ARCtrl_DataMap__DataMap_tryFromReadContractForStudy_Static", "ARCtrl_DataMap__DataMap_ToCreateContractForWorkflow_Z721C83C5", "ARCtrl_DataMap__DataMap_ToUpdateContractForWorkflow_Z721C83C5", "ARCtrl_DataMap__DataMap_ToDeleteContractForWorkflow_Z721C83C5", "ARCtrl_DataMap__DataMap_toDeleteContractForWorkflow_Static_Z721C83C5", "ARCtrl_DataMap__DataMap_toUpdateContractForWorkflow_Static_Z721C83C5", "ARCtrl_DataMap__DataMap_tryFromReadContractForWorkflow_Static", "ARCtrl_DataMap__DataMap_ToCreateContractForRun_Z721C83C5", "ARCtrl_DataMap__DataMap_ToUpdateContractForRun_Z721C83C5", "ARCtrl_DataMap__DataMap_ToDeleteContractForRun_Z721C83C5", "ARCtrl_DataMap__DataMap_toDeleteContractForRun_Static_Z721C83C5", "ARCtrl_DataMap__DataMap_toUpdateContractForRun_Static_Z721C83C5", "ARCtrl_DataMap__DataMap_tryFromReadContractForRun_Static"]




from __future__ import annotations

import array
from abc import abstractmethod
from collections.abc import Callable, Iterable, MutableSequence
from typing import (
    Any,
    Generic,
    TypeVar,
    cast,
)

from .util import Array, IComparable, compare


_T = TypeVar("_T")


class FSharpRef(Generic[_T]):
    __slots__ = "getter", "setter"

    def __init__(
        self,
        contents_or_getter: None | (_T | Callable[[], _T]),
        setter: Callable[[_T], None] | None = None,
    ) -> None:
        contents = cast(_T, contents_or_getter)

        def set_contents(value: _T):
            nonlocal contents
            contents = value

        if callable(setter):
            self.getter = cast(Callable[[], _T], contents_or_getter)
            self.setter = setter
        else:
            self.getter = lambda: contents
            self.setter = set_contents

    @property
    def contents(self) -> _T:
        return self.getter()

    @contents.setter
    def contents(self, v: _T) -> None:
        self.setter(v)


class Union(IComparable):
    __slots__ = "tag", "fields"

    def __init__(self):
        self.tag: int
        self.fields: Array[Any] = []

    @staticmethod
    @abstractmethod
    def cases() -> list[str]: ...

    @property
    def name(self) -> str:
        return self.cases()[self.tag]

    def __str__(self) -> str:
        if not len(self.fields):
            return self.name

        def to_string(value: Any) -> str:
            if isinstance(value, str):
                return f'"{value}"'
            return str(value)

        fields = ""
        with_parens = True
        if len(self.fields) == 1:
            field = to_string(self.fields[0])
            with_parens = field.find(" ") >= 0
            fields = field
        else:
            fields = ", ".join(map(to_string, self.fields))

        return self.name + (" (" if with_parens else " ") + fields + (")" if with_parens else "")

    def __repr__(self) -> str:
        return str(self)

    def __hash__(self) -> int:
        hashes = map(hash, self.fields)
        return hash((hash(self.tag), *hashes))

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True

        if not isinstance(other, Union):
            return False

        # Different objects are not equal even with same structure
        if self.__class__ != other.__class__:
            return False

        if self.tag == other.tag:
            return self.fields == other.fields

        return False

    def __cmp__(self, __other: Any) -> int:
        if self < __other:
            return -1
        elif self == __other:
            return 0
        return 1

    def __lt__(self, other: Any) -> bool:
        if self.tag == other.tag:
            return True if compare(self.fields, other.fields) < 0 else False

        return self.tag < other.tag


def record_compare_to(self: Record, other: Record) -> int:
    if self is other:
        return 0

    def compare_values(self_value: Any, other_value: Any) -> int:
        match (self_value, other_value):
            case (None, None):
                return 0
            case (None, _):
                return -1
            case (_, None):
                return 1
            # Check for custom equality
            case (self_value, other_value) if self_value == other_value:
                return 0
            case (self_value, other_value) if self_value < other_value:
                return -1
            case (self_value, other_value) if self_value > other_value:
                return 1
            case _:
                return 0

    if hasattr(self, "__dict__") and self.__dict__:
        for name in self.__dict__.keys():
            result = compare_values(self.__dict__[name], other.__dict__[name])
            if result != 0:
                return result

    elif hasattr(self, "__slots__") and self.__slots__:
        for name in self.__slots__:
            result = compare_values(getattr(self, name), getattr(other, name))
            if result != 0:
                return result

    return 0


def record_equals(self: _T, other: _T) -> bool:
    if self is other:
        return True

    if self.__class__ != other.__class__:
        return False

    if isinstance(self, Record) and isinstance(other, Record):
        return record_compare_to(self, other) == 0

    return self == other


def record_to_string(self: Record) -> str:
    if hasattr(self, "__slots__"):
        return "{ " + "\n  ".join(map(lambda slot: slot + " = " + str(getattr(self, slot)), self.__slots__)) + " }"
    else:
        return "{ " + "\n  ".join(map(lambda kv: kv[0] + " = " + str(kv[1]), self.__dict__.items())) + " }"


def record_get_hashcode(self: Record) -> int:
    slots = type(self).__slots__
    return hash(tuple(getattr(self, fixed_field) for fixed_field in slots))


class Record(IComparable):
    __slots__: list[str]

    def GetHashCode(self) -> int:
        return record_get_hashcode(self)

    def Equals(self, other: Record) -> bool:
        return record_equals(self, other)

    def __cmp__(self, other: Record) -> int:
        return record_compare_to(self, other)

    def __str__(self) -> str:
        return record_to_string(self)

    def __repr__(self) -> str:
        return str(self)

    def __lt__(self, other: Any) -> bool:
        return True if self.__cmp__(other) == -1 else False

    def __eq__(self, other: Any) -> bool:
        return self.Equals(other)

    def __hash__(self) -> int:
        return record_get_hashcode(self)


class Attribute: ...


def seq_to_string(self: Iterable[Any]) -> str:
    str = "["

    for count, x in enumerate(self):
        if count == 0:
            str += to_string(x)

        elif count == 100:
            str += "; ..."
            break

        else:
            str += "; " + to_string(x)

    return str + "]"


def to_string(x: Iterable[Any] | Any, call_stack: int = 0) -> str:
    if x is not None:
        if isinstance(x, float) and int(x) == x:
            return str(int(x))

        if isinstance(x, bool):
            return str(x).lower()

        if isinstance(x, Iterable) and not hasattr(x, "__str__"):
            return seq_to_string(x)

    return str(x)


class FSharpException(Exception, IComparable):
    def __init__(self) -> None:
        self.Data0: Any = None

    def __str__(self) -> str:
        return record_to_string(self)  # type: ignore

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True

        if other is None:
            return False

        return self.Data0 == other.Data0

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, FSharpException):
            return False

        if self.Data0:
            if other.Data0:
                return self.Data0 < other.Data0
            else:
                return False

        elif not self.Data0:
            if other.Data0:
                return False
            else:
                return True

        return super().__lt__(other)

    def __hash__(self) -> int:
        return hash(self.Data0)

    def GetHashCode(self) -> int:
        return hash(self)

    def Equals(self, other: FSharpException):
        return record_equals(self, other)

    def CompareTo(self, other: FSharpException):
        return compare(self.Data0, other.Data0)


class char(int):
    __slots__ = ()


class int8(int):
    __slots__ = ()


class int16(int):
    __slots__ = ()


class int32(int):
    __slots__ = ()


class int64(int):
    __slots__ = ()


class uint8(int):
    __slots__ = ()


class uint16(int):
    __slots__ = ()


class uint32(int):
    __slots__ = ()


class uint64(int):
    __slots__ = ()


class float32(float):
    __slots__ = ()


float = float  # use native float for float64


def Int8Array(lst: list[int]) -> MutableSequence[int]:
    return array.array("b", lst)


def Uint8Array(lst: list[int]) -> MutableSequence[int]:
    return bytearray(lst)


def Int16Array(lst: list[int]) -> MutableSequence[int]:
    return array.array("h", lst)


def Uint16Array(lst: list[int]) -> MutableSequence[int]:
    return array.array("H", lst)


def Int32Array(lst: list[int]) -> MutableSequence[int]:
    return array.array("i", lst)


def Uint32Array(lst: list[int]) -> MutableSequence[int]:
    return array.array("I", lst)


def Float32Array(lst: list[float]) -> MutableSequence[float]:
    return array.array("f", lst)


def Float64Array(lst: list[float]) -> MutableSequence[float]:
    return array.array("d", lst)


def is_exception(x: Any):
    return isinstance(x, Exception)


__all__ = [
    "Attribute",
    "Array",
    "is_exception",
    "char",
    "int8",
    "uint8",
    "int16",
    "uint16",
    "int32",
    "uint32",
    "int64",
    "uint64",
    "Int8Array",
    "Uint8Array",
    "Int16Array",
    "Uint16Array",
    "Int32Array",
    "Uint32Array",
    "Float32Array",
    "Float64Array",
    "FSharpException",
    "FSharpRef",
    "Record",
    "seq_to_string",
    "to_string",
    "Union",
]
